module Main where

import           Control.DeepSeq     (deepseq)
import           Control.Monad       (join)
import           Data.Monoid         ((<>))
import qualified Data.Text.Lazy      as T
import qualified Data.Text.Lazy.IO   as TIO
import           Lib
import           Options.Applicative
import           System.Exit         (die, exitSuccess)
import           System.IO           (hFlush, stdout)

-- Adapted with permission from https://bit.ly/35WGzyB
main :: IO ()
main = join . customExecParser (prefs showHelpOnError) $
  info (helper <*> parser)
    (fullDesc <>
     header
       "Parallelized N-Gram Language Modeling with Stupid Backoff for Text Generation" <>
     progDesc ("Build a language model from a given corpus, then use it to assign" ++ 
     " probability scores to input lines, or to auto-complete them" ++
     " (similar to your smartphone keyboard)."))
  where
    parser :: Parser (IO ())
    parser =
      work <$>
      strOption
        (long "mode" <> short 'm' <> metavar "MODE" 
        <> help "Operating mode (buildOnly, score, complete)") <*>
      strOption
        (long "corpus-file" <> short 'f' <> metavar "FILENAME" 
        <> help "Path to corpus file") <*>
      option auto
        (long "ngrams" <> short 'n' <> metavar "NUMBER" <> help "Size of n-gram" 
        <> value 3 <> showDefault) <*>
      option auto
        (long "ndocs" <> metavar "NUMBER" <> value (-1) <> showDefault <>
         help "Number of documents to parse (-1 to parse all)") <*>
      strOption
        (long "doc-separator" <> short 's' <> metavar "SEPARATOR" <>
         value "---END.OF.DOCUMENT---" <> showDefault <>
         help "Line that separates documents in corpus") <*>
      option auto
        (long "freq-threshold" <> short 't' <> metavar "NUMBER" <> value 0 <> showDefault
         <> help "Prune all n-grams that occur fewer than k times") <*>
      strOption
        (long "complete-mode" <> metavar "MODE" <> value "beamSearch" <> showDefault <>
         help "Operating mode if --mode=complete (beamSearch, greedy, random)")

work :: String -> String -> Int -> Int -> String -> Int -> String -> IO ()
work mode corpusFile n nDocs docSep thresh completeMode = do
  corpusText <- TIO.readFile corpusFile
  let untokCorpus = 
        splitIntoDocs (T.pack docSep) $ filter (/= T.empty) $ T.lines corpusText
      corpus = par' $ map (map tokenize) untokCorpus
      maybePartialCorpus = if nDocs < 0 then corpus else take nDocs corpus
      nGramTries = triesFrom n thresh maybePartialCorpus
  if mode == "buildOnly" then do
    putStrLn "Building n-gram model..."
    putStrLn $ nGramTries `deepseq` "Built."
    exitSuccess
  else if mode == "score" then scoreLoop nGramTries n
  else if mode == "complete" then completeLoop nGramTries n completeMode
  else die $ "Mode must be one of [score, complete]. You gave " ++ mode ++ "."

promptLoop :: String -> (String -> IO String) -> IO ()
promptLoop prompt f = do
  putStr prompt
  hFlush stdout
  line <- getLine
  if null line
    then return ()
  else do
    out <- f line
    putStrLn out
    promptLoop prompt f

scoreLoop :: [Trie] -> Int -> IO ()
scoreLoop tries n = promptLoop "Score> " $ return . show . lineScore tries n . T.pack

completeLoop :: [Trie] -> Int -> String -> IO ()
completeLoop tries n mode = promptLoop "Complete> " $ lineComplete mode tries n . T.pack
